#include <stdarg.h>
#include <unistd.h>
#include <string.h>
#include <ncurses.h>
#include <time.h>

#include "iface.h"
#include "thread.h"

using namespace std;

const char banner[]="[proSilla " VERSION " - download accelerator for ssh] <anszom@v-lo.krakow.pl>";

/***/

int iface::prepareinfo(thread * th, int n, char * buf, fsize_t size)
{
fsize_t bytez, initbytez, max;
//int transfer=-1;
int ok;
int l=0;	

	if(th){
	bytez = th->bytes;
	initbytez = th->initbytes;
	max = th->max;
	ok=th->status;
	l+=snprintf(buf+l,size-l,"%2d. ",n+1);
//		if(th && th->bytes > th->initbytes && th->transfer>=0)
//		transfer=th->transfer;
	}else{
	bytez = fle.totalbytes;
	initbytez = fle.totinitbytes;
	max = fle.size;
	ok=(bytez<max)?1:2;
	l+=snprintf(buf+l,size-l,"Tot.");
//		if(thread::fulltransfer>=0) transfer=thread::fulltransfer;
	}

	if(bytez < initbytez) bytez = initbytez;	// ??
	
	switch(ok){
	case 0:
	case 1:
	l+=snprintf(buf+l,size-l,"%8lu kb (%2d%%)", bytez>>10,
					(int)(100*(bytez>>10)/(max>>10)));
		
		if(ok==0){
		l+=snprintf(buf+l,size-l," CONNECTING");
		break;
		}
				
		if(now>started)l+=snprintf(buf+l,size-l," %6.2fkb/s",
			(double)(bytez - initbytez)/(1024*(now-started)));
		
//		if(transfer>=0)	l+=snprintf(buf+l,size-l," cur: %6.2fkb/s",
//			(double)transfer/(1024*5));

		if(now>started){
		int eta_s=eta(bytez - initbytez, max - initbytez);
		l+=snprintf(buf+l,size-l," ETA %02d:%02d:%02d",
					(eta_s/3600),(eta_s/60)%60,eta_s%60);
		}
	break;
	
	case 2:
	l+=snprintf(buf+l,size-l,"FINISHED");
		if(now>started)
		l+=snprintf(buf+l,size-l," %6.2fkb/s",
			(double)(bytez - initbytez)/(1024*(now-started)));
	}
	
	if(!th && now>started)
	l+=snprintf(buf+l,size-l," elapsed: %02d:%02d:%02d",
		(int)(now-started)/3600,(int)((now-started)/60)%60, 
		(int)(now-started)%60);
return l;
}

/***/

void iface::printerror(int fd)
{
char buf[1024]; int tl=0;

	do{
	int l=read(fd, buf+tl,1);
		if(l != 1) throw "read";
	}while(buf[tl++] != '\n');

	if(buf[tl-2]=='\r') buf[tl-2]=' ';
buf[tl]=0;

struct tm * tm=localtime(&now);
msg(2, "[%2d:%02d:%02d] %s", tm->tm_hour, tm->tm_min, tm->tm_sec, buf);
}

/***/

void iface::get_str(char * prompt, char * buf, int n, bool pass)
{
	if(pass)
	strncpy(buf,getpass(prompt),n);
	else{
	printf("%s", prompt); fflush(stdout);
	fgets(buf, n, stdin);
	}
}

/***/

void simple::refresh()
{
fflush(stdout);
}

/***/

void simple::updateinfo(thread * th, int n)
{
char buf[1024];
static int oldl;
int l;
	if(th) return;
l=prepareinfo(NULL,0,buf,sizeof(buf));
	if(oldl>l) memset(buf+l,' ',oldl-l);
buf[oldl]=0;
oldl=l;
printf("\r%s",buf); refresh();
printedstatus=true;
}

/***/

void simple::msg(int l, char *fmt, ...)
{
va_list ap;
va_start(ap,fmt);
	if(printedstatus) printf("\n");
vprintf(fmt,ap);
va_end(ap);
}

/***/

void simple::eatkey()
{
char c;
printf("Press enter...\n");
read(0,&c,1);
}

/***/

simple::simple()
{
printedstatus=false;
printf("%s\n",banner);
}

/***/

ncurses::ncurses()
{
initscr();
crmode();
nonl();
noecho();
Wmain = newwin(LINES - cons_h, 0, 0, 0);
Wcons = newwin(cons_h, 0, LINES - cons_h - 1, 0);

scrollok(Wcons, true);
wclear(Wmain);
wclear(Wcons);
wmove(Wmain, 0, 0);
wprintw(Wmain, "%s",banner);
}

/***/

ncurses::~ncurses()
{
nocrmode();
nl();
echo();
endwin();
}

/***/

void ncurses::updateinfo (thread * th, int n)
{
char buf[1024];
prepareinfo(th,n,buf,sizeof(buf));
	if(th){
		if(4+n > COLS - 3 - cons_h)return;
	wmove(Wmain, 4+n,0);
	}else wmove(Wmain, 4+fle.parts,0);
wclrtoeol(Wmain);
wprintw(Wmain, "%s",buf);
}

/***/

void ncurses::msg(int l,char *fmt,...)
{
va_list ap;
va_start(ap,fmt);
	if(l<2){
	wmove(Wmain, l+1, 0);
	wclrtoeol(Wmain);
	vwprintw(Wmain, fmt, ap);
	}else{
	vwprintw(Wcons, fmt, ap);
	}
va_end(ap);
}

/***/

void ncurses::refresh()
{
wnoutrefresh(Wcons);
wnoutrefresh(Wmain);
doupdate();
}

/***/

void ncurses::eatkey()
{
scrollok(Wcons, false);
wmove(Wcons, cons_h-1, 0);
wprintw(Wcons, "Press any key...");
wrefresh(Wcons);
wgetch(Wcons);
wmove(Wcons, cons_h-1, 0);
wclrtoeol(Wcons);
scrollok(Wcons, true);
}

/***/

void ncurses::get_str(char * prompt, char * buf, int n, bool pass)
{
msg(1, "%s", prompt);
refresh();
	if(!pass) echo();
wgetnstr(Wmain, buf, n);
	if(!pass) noecho();

msg(1, "");
refresh();
}

/***/
