/*
    proSilla - transfer accelerator working over ssh
    Copyright (C) 2002  Andrzej Szombierski <anszom@v-lo.krakow.pl>

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
*/


#include <errno.h>
#include <string.h>
#include <iostream>
#include <fstream>
#include <unistd.h>
#include <signal.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <stdlib.h>
#include <sstream>
#include <sys/select.h>
#include <fcntl.h>
#include <iomanip>
#include <time.h>
#include <pty.h>
#include <getopt.h>
#include <sys/file.h>

#include "thread.h"
#include "iface.h"

#ifdef WITH_GTK
#include "gtk_iface.h"
#define IFGTK(a)	a
#else
#define IFGTK(a)
#endif

time_t started;
time_t now;

iface * Iface;

using namespace std;

/***/

void usage(char * name)
{
cerr << "Usage: " << name << " [options] [user@]host file" <<
endl << "   or: " << name << " -i file.psinfo [file2.psinfo...]" <<
endl << 
endl << "Options:" <<
endl << "-b          Go to background after logging in, implies -q" <<
endl << "-c          Continue a simple download (scp/wget/... *NOT* proSilla)" <<
endl << "-f          Ignore timestamp/size mismatch" <<
endl << "-k n        Use n connections (default=4)" << 
endl << "-p[proserv] Use proserv instead of /bin/dd on remote host" << 
endl << "            If no argument given, assume ./bin/proserv" <<
endl << "-s arg      Pass 'arg' to ssh" <<
endl << "            example: -s -cblowfish, or -s -c -s blowfish" << 
endl << "-C          Use ssh compression (equivalent to -s -C)" <<
endl <<
IFGTK(endl << "-g          Use the gtk+ interface" <<)
endl << "-N          Don't use the ncurses interface" << 
endl << "-n          Don't wait for a keystroke before exiting" <<
endl << "-q          Don't show anything, just download" <<
endl <<
endl << "-i          Don't download anything, just print progress information" <<
endl << "            from a .psinfo file" <<
endl <<
endl << "Note: Do NOT use this program on a fast connection (LAN or something)." <<
endl << "It will only slow things down :(" <<
endl;
}

/***/

void sighandler(int sig)
{
fle.writepsinfo();
delete Iface;
exit(1);
}

/***/

void chldhandler(int sig)
{
wait(NULL);
}

/***/

void hup(int sig)
{
delete Iface;
Iface = new iface();
	if(fork()) _exit(0);
signal(SIGHUP, SIG_IGN);
}

/***/

int main(int ac, char ** av)
{
char c;
int iftype=2;
bool need_init=false;

bool bg=false;
bool nowait=false;

fle.parts=4;
fle.recover=false;
fle.force=false;

bool printinfo=false;

#ifdef WITH_GTK
char * tmp=strrchr(av[0], '/');
	if(!tmp) tmp=av[0];
	else tmp++;

	if(*tmp=='g') iftype=3;
#endif

	while((c=getopt(ac,av,"bck:p::qs:ngNfiC"))!=EOF) switch(c){
		case 'b':
		iftype=0;
		bg=true;
		break;

		case 'c':
		fle.recover=true;
		break;
		
		case 'k':
		fle.parts=atoi(optarg);
			if(fle.parts < 1 || fle.parts > 20){
			cerr << "Nonsense number of connections" << endl;
			return 1;
			}
		break;

		case 'n':
		nowait=true;
		break;
		
		case 'p':
			if(optarg) thread::proserv=optarg;
			else thread::proserv="./bin/proserv";
		break;
		
		case 'q':
		iftype=0;
		break;

		case 's':
		fajl::sshargs.push_back(optarg);
		break;

		case 'g':
#ifdef WITH_GTK
		iftype=3;
		break;
#else
		cerr << "Gtk+ support not compiled in!" << endl;
		return 1;
#endif
		
		case 'N':
		iftype=1;
		break;

		case 'C':
		thread::compress=true;
		break;

		case 'f':
		fle.force=true;
		break;

		case 'i':
		printinfo=true;
		break;
		
		default:
		usage(av[0]);
		return 1;
	}

signal(SIGPIPE, SIG_IGN);
signal(SIGCHLD, chldhandler);
signal(SIGHUP, hup);

	if(printinfo){
		if(ac-optind<1) {
		usage(av[0]);
		return 1;
		}
		
	Iface = new simple();
	nowait=true;
		for(int i=optind;av[i];i++){
			try{
			fajl::printinfo(av[i]);
			
			}catch(const char * s){
				
			Iface->msg(2,"%s: %s\n",s,strerror(errno));
			
			}catch(int){
			}
		}

	return 0;
	}

char *host=NULL, *file=NULL;
	if(ac-optind!=2){
		if(iftype==3 && ac-optind==0) need_init=true;
		else{
		usage(av[0]);
		return 1;
		}
		
	}else{
	host=av[optind];
	file=av[optind+1];
	}


	if(bg && iftype){
	cerr << "You can't use the -b (background) option with -N or -g"<< endl;
	return 1;
	}

	try{
		switch(iftype){
		case 0:	Iface = new iface();	break;
		case 1: Iface = new simple();	break;
		case 2: Iface = new ncurses();	break;
		IFGTK(case 3: Iface = new gtk(&ac, &av); break;)
		}
		
		if(need_init) if(!Iface->setup(host, file)) throw 0;
	thread::init();

	now=time(NULL);
	
	fle.setup(file, host);
	
	threads = new thread[fle.parts];

	Iface->msg(0, "*** downloading %s:%s\n", host, file);
	Iface->msg(1, "Getting file info...\n");
	Iface->refresh();
	
	fle.get_info();
	
	// we're authenticated so we can continue alone...
		if(bg){
			if(fork()) _exit(0);
		setsid();
		}

	now=started=time(NULL);
	
		for(int i=0;i<fle.parts;i++){
			if(threads[i].status != thread::finished)threads[i].connect();
		Iface->updateinfo(threads+i,i);
		}
	Iface->updateinfo(NULL,0);
	Iface->refresh();

	time_t lastupd=0;
	//time_t lastrecalc=0;FIXME

	signal(SIGINT, sighandler);
	signal(SIGTERM, sighandler);
	
		while(fle.totalbytes < fle.size){
		fd_set zfds=thread::fds;
		
		timeval tv;
		tv.tv_sec=2;
		tv.tv_usec=0;
	
		int n=select(thread::maxfd, &zfds, NULL, NULL, &tv);
		now=time(NULL);
			if(fle.totalbytes == fle.totinitbytes) started=now;
	
			if(n<0){
				if(errno==EINTR)continue;
			throw "select";
			}
			
			if(FD_ISSET(thread::errpipe[0],&zfds)){
			Iface->printerror(thread::errpipe[0]);
			n--;
			}

			for(int i=0;i<fle.parts;i++){
				if(threads[i].status != thread::finished)
				if(FD_ISSET(threads[i].masterpty,&zfds)){
				write(threads[i].masterpty,thread::password,
						strlen(thread::password));
				write(threads[i].masterpty,"\n",1);
				n--;
				char c;
					do{
					read(threads[i].masterpty,&c,1);
					}while(c!='\n');
				FD_CLR(threads[i].masterpty,&thread::fds);
				}
			}
		
			for(int i=0;i<fle.parts;i++){
				if(threads[i].status != thread::finished)
				if(FD_ISSET(threads[i].pipe_fd, &zfds)){
				threads[i].receive();
				Iface->updateinfo(threads+i,i);
				n--;
				}
			}
			
/*FIXME			if(lastrecalc+5<now){
			lastrecalc=now;
			thread::fulltransfer=0;
				for(int i=0;i<fle.parts;i++)
				if(threads[i].status != thread::transferring) threads[i].recalc();
			}*/

		Iface->updateinfo(NULL,0);
			if(lastupd<now){
			lastupd=now;
				if(fle.parts > 1)
				for(int i=0;i<fle.parts;i++)
				if(threads[i].status != thread::finished)
				Iface->updateinfo(threads+i,i);
			fle.writepsinfo();
			}
			
			if(n)Iface->msg(2,"Shit! the programmer forgot to check something (%d times!)\n",n);
		Iface->refresh();
		}

	fle.rmpsinfo();

	Iface->msg(1, "Download complete!\n");
		if(!nowait)Iface->eatkey();
	
	delete Iface;
	return 0;
	
	}catch(const char * s){
	Iface->msg(2,"%s: %s\n",s,strerror(errno));

	}catch(int){
	}

	if(!nowait)Iface->eatkey();
delete Iface;
return 1;
}
