#include <unistd.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <sys/wait.h>
#include <sys/select.h>
#include <sys/file.h>
#include <termios.h>
#include <fcntl.h>

#include <errno.h>
#include <string.h>
#include <stdlib.h>
#include <time.h>

#include <pty.h>

//#include <fstream.h>
#include <iomanip>

#include <vector>
#include <algorithm>

#include "thread.h"
#include "iface.h"

// configuration
bool		thread::compress=false;
char		*thread::proserv=NULL;

// select() stuff
int		thread::maxfd;
fd_set		thread::fds;

// sshs' stderr and password
int		thread::errpipe[2];
char		thread::password[];

//int		thread::fulltransfer=-1;

vector<char*> fajl::sshargs;

/***/

thread * threads;
fajl fle;

/***/

int fajl::spawn_ssh(int *pty, int stout, int sterr, char *cmd, bool compress)
{
char ** argv;
int n=0;
int pid;
argv=new char*[6+sshargs.size()];
	// ssh -n user@host [-C] command <sshargs> NULL

	if(pty)
	pid=forkpty(pty,NULL,NULL,NULL);
	else
	pid=fork();

	if(pid){
	close(stout);
	return pid;
	}
	
argv[n++]="ssh";
argv[n++]="-n";
argv[n++]=fle.source;
	if(compress) argv[n++]="-C";
copy(sshargs.begin(), sshargs.end(), argv+n);
n+=sshargs.size();

argv[n++]=cmd;
argv[n++]=NULL;

	if(stout >= 0) dup2(stout,1);
	if(sterr >= 0) dup2(sterr,2);
		for(int i=3;i<64;i++)close(i);
execvp("ssh", argv);
throw "execvp";
}

/***/

void thread::connect()
{
int pajp[2];
	if(pipe(pajp)<0) throw "pipe";
char cmdbuf[1024];
status = connecting;

	if(proserv){
	ignorebytes=0;
	snprintf(cmdbuf,sizeof(cmdbuf), "%s %s %lu %lu", proserv, fle.mnglname,
			offset + bytes, max - bytes);
	}else{
	fsize_t start = offset+bytes;
	ignorebytes = start % 1024;
	fsize_t len = max - bytes +ignorebytes;
	fsize_t z = len % 1024;
		if(z) len+=(1024-z);
	snprintf(cmdbuf,sizeof(cmdbuf),
		"/bin/dd if=%s bs=1024 skip=%lu count=%lu 2>/dev/null",
		fle.mnglname, start>>10, len>>10);
	}

fle.spawn_ssh(&masterpty, pajp[1], errpipe[1], cmdbuf, compress);

pipe_fd=pajp[0];

	if(pajp[0]>=maxfd) maxfd=pajp[0]+1;
	if(masterpty>=maxfd) maxfd=masterpty+1;
FD_SET(pajp[0], &fds);
FD_SET(masterpty, &fds);
//transfer=-1;
}

/***/

void thread::remove()
{
FD_CLR(pipe_fd,&fds);
FD_CLR(masterpty,&fds);
close(pipe_fd);
close(masterpty);
}

/***/

void thread::receive()
{
fsize_t l;
char buf[4096];
char *b=buf;
l=read(pipe_fd, buf, sizeof(buf));
	if(l<0) throw "read";
	
	if(l==0){
	remove();
		if(!is_finished()){
		// (TODO: implement reconnect delay?)
		connect();	// reconnect
		}else{
		status=finished;
		}
	return;
	}
	
	if(ignorebytes){	// should we ignore something
		if(l>ignorebytes){
		l-=ignorebytes;	// only a part of this chunk
		b+=ignorebytes;
		ignorebytes=0;
		}else{		// whole chunk should be ignored
		ignorebytes-=l;
		return;
		}
	}		
	
	// cut off the trailing junk
	if(l > max - bytes) l = max - bytes;

status = transferring;

lseek(fle.fd,bytes + offset,SEEK_SET);
bytes+=l;	// adjust the counters
fle.totalbytes+=l;
unsigned int n=write(fle.fd,b,l);
	if(n<=0)throw "write()";

	if(n!=l){
	Iface->msg(2, "Short write!\n");
	throw 0;
	}
	
	if(is_finished()){
	status = finished;
	remove();
	}
}

/***/

void thread::init()
{
	if(pipe(errpipe))throw "pipe";
	if(errpipe[0] >= maxfd)
	maxfd=errpipe[0]+1;
	
FD_ZERO(&fds);	
FD_SET(errpipe[0],&fds);
}

/*****************************************************************************/

void fajl::setup(char * nm, char * sr)
{
name=nm; source=sr;
unsigned int z=0;
mnglname[z++]='\"';

	// FIXME: do we always want to escape the strings?
	// imagine: prosilla host *.tgz

	// "`\ += '\'
	for(char *c=name;*c && z<sizeof(mnglname)-1;c++){
		if(strchr("\\\"`",*c)&&(z<sizeof(mnglname)-2))
		mnglname[z++]='\\';
	mnglname[z++]=*c;
	}

mnglname[z++]='\"';
mnglname[z++]=0;

localname=strrchr(name,'/');
	if(localname)localname++;
	else localname=name;

strncpy(psinfo,localname,sizeof(psinfo)-8);
psinfo[sizeof(psinfo)-8]=0;
strcat(psinfo,".psinfo");
}
	
/***/

void fajl::initpsinfo()
{
psifd=open(psinfo,O_RDWR|O_CREAT,0600);
	if(psifd<0)throw "open(.psinfo)";
	
psihdr psih;
	if(::flock(psifd,LOCK_EX | LOCK_NB)<0){	// no other prosilla running?
	Iface->msg(2, "Can't lock the .psinfo file (is it used by another proSilla)?\n");
	throw 0;
	}
	
	if(read(psifd,&psih,sizeof(psihdr))==sizeof(psihdr)){
		if(recover){
		Iface->msg(2,"ERROR: %s was found.\n", psinfo);
		Iface->msg(2,"If you are trying to resume a proSilla download, don't use -c\n");
		throw 0;
		}

		if(force){
		memcpy(psih.stamp,stamp,14);
		psih.size=size;
		}
		
		if(memcmp(stamp,psih.stamp,14)){
		Iface->msg(2, "The timestamp is different!\n");
		Iface->msg(2, "If you are sure this is the same file, use -f\n");
		throw 0;
		}
			
		if(parts!=psih.parts){
		Iface->msg(2, "Set parts to %d!\n",psih.parts);
		throw 0;
		}
			
		if(size!=psih.size){
		Iface->msg(2, "Original file size is different!\n");
		throw 0;
		}
	totalbytes=0;
		
		for(int i=0;i<parts;i++){
		read(psifd,&threads[i].bytes,sizeof(int));
		threads[i].initbytes=threads[i].bytes;
		totalbytes+=threads[i].bytes;
		}
	
	totinitbytes=totalbytes;
	}else{
	lseek(psifd,0,SEEK_SET);

		if(recover){
		struct stat st;
		fstat(fd, &st);
		fsize_t stsize=st.st_size;
		totinitbytes=totalbytes=stsize;
			for(int i=0;i<parts;i++){
			threads[i].initbytes=threads[i].bytes=
				(stsize > block)?block:stsize;
			stsize-=threads[i].bytes;
			}
		}else{
			for(int i=0;i<parts;i++)
			threads[i].bytes=threads[i].initbytes=0;
		}
	
	memcpy(psih.stamp, stamp, 14);
	psih.parts=parts;
	psih.size=size;
	write(psifd,&psih,sizeof(psihdr));
	}
	
	for(int i=0;i<parts;i++){
	threads[i].offset=i*block;  
	threads[i].max=(i+1==parts)?(size % block) : block;
		if(threads[i].max==threads[i].bytes)
		threads[i].status = thread::finished;
		else
		threads[i].status = thread::connecting;
	}
}

/***/

void fajl::get_info()
{
char buf[1000];
snprintf(buf, sizeof(buf),
		"/bin/date -r %s '+%%Y%%m%%d%%H%%M%%S'; /bin/ls -lL %s",
		mnglname, mnglname);

int output[2];
	if(pipe(output)) throw "pipe" ;

int masterpty;
spawn_ssh(&masterpty, output[1], thread::errpipe[1], buf);

	try{
	int m=output[0];
		if(thread::errpipe[0]>m) m=thread::errpipe[0];
		if(masterpty>m) m=masterpty;
	
	fd_set tmp;
	int inbuf=0;

		do{
		FD_ZERO(&tmp);
		FD_SET(output[0],&tmp);
		FD_SET(thread::errpipe[0],&tmp);
		FD_SET(masterpty,&tmp);
			if(select(m+1,&tmp,NULL,NULL,NULL)<0){
				if(errno==EINTR)continue;
			throw "select";
			}
			
			if(FD_ISSET(thread::errpipe[0],&tmp))
			Iface->printerror(thread::errpipe[0]);

			if(FD_ISSET(masterpty, &tmp)){
				if(inbuf+1 == sizeof(buf)){
				Iface->msg(2, "ssh client has gone crazy!\n");
				inbuf=0;
				}

			int n=read(masterpty, buf+inbuf, sizeof(buf)-inbuf-1);
				if(n<=0) throw 1;
			int j=0;
				for(int i=0;i<n;i++){
				buf[inbuf+j]=buf[inbuf+i];
					if(buf[inbuf+i] != '\r') j++;
				}
			inbuf+=j;
			
				// just line-splitting code...
				while(inbuf){
				buf[inbuf]=0;
				char * c= strchr(buf, '\n');
					if(!c) break;
				*c=0;
				Iface->msg (2, "%s\n", buf);
				Iface->refresh ();
				memmove(buf, c+1, inbuf-(c-buf+1));
				inbuf-=(c-buf+1);
				}


		// probably "The authenticity of host <..> can't be established 
				if(strncmp("(yes/no)? ", buf+inbuf-10, 10)==0
				|| strncmp("'yes' or 'no': ", buf+inbuf-15, 15)==0){
				char resp[100];
				buf[inbuf]=0;	// prompt string
				Iface->get_str(buf, resp, sizeof(resp));
				
				write(masterpty, resp, strlen(resp));
				write(masterpty,"\n",1);
				inbuf=0;

				// it might be a password prompt...
				}else if(strncmp(": ", buf+inbuf-2, 2)==0){
				struct termios tio;

					if(tcgetattr(masterpty, &tio))
					throw "tcgetattr";
	
					// noecho mode only
					if(!(tio.c_lflag & ECHO)){
					buf[inbuf]=0;	// prompt string

					Iface->refresh ();
					Iface->get_str (buf, thread::password,
						sizeof(thread::password), true);
		
					// inject the password
					write(masterpty,thread::password,
						strlen(thread::password));
					write(masterpty,"\n",1);
				
						// ignore the rest...
/*						do{
						read(masterpty, buf, 1);
						}while(buf[0] != '\n');*/
					inbuf=0;
					}
				}
			
			}
	
		}while(!FD_ISSET(output[0],&tmp));

		{
		FILE *fpipe=fdopen(output[0], "r");
		int ok=1;
			if(!fgets(buf, sizeof(buf), fpipe)) ok=0;
			else{
				if(buf[0]) buf[strlen(buf)-1]=0;
			strncpy(stamp, buf, 15);
			stamp[14]=0;
			}
			
			if(!fgets(buf, sizeof(buf), fpipe)) ok=0;
			else buf[strlen(buf)-1]=0;

		// this used to work in gcc 2.95
//		if(! ifstream(output[0]).getline(stamp, 15)
//			.getline(buf, sizeof(buf))){
			if(!ok){
			Iface->msg (2, "Error reading from pipe\n");
			fclose(fpipe);
			throw 0;
			}
		}
	
	char * endp;
	char * token=buf;
	// extract the fifth token - file size (or device major number :)
		for(int n=0;n<4;n++){
			while(*token!=' ' && *token!='\t') token++;
			while(*token==' ' || *token=='\t') token++;
		}

		for(endp=token;*endp!=' ' && *endp!='\t';endp++);
	*endp=0;
	size=strtol(token, &endp, 10);
	
		if(*endp){
		Iface->msg (2, "Incorrect file size string: %s\n", buf);
		throw 0;
		}

	close(masterpty);	// This should kill the client
	close(output[0]);
	
	block=(size/parts)+1;
	
	Iface->msg(1, "Total file size: %lu kb, %d blocks of %lu bytes (+/-1)\n", size/1024, parts, block);
		
	fd=open(localname, O_WRONLY|O_CREAT,0600);
		if(fd<0) throw "open(file)";
		
	initpsinfo();
	
	Iface->refresh();
	}catch(...){
	close(masterpty);
	close(output[0]);
	throw;
	}
// FIXME: do we need wait(NULL);?
}

/***/

void fajl::writepsinfo()
{
lseek(psifd,sizeof(psihdr),SEEK_SET);
	for(int i=0;i<parts;i++)
	write(psifd,(char*)&threads[i].bytes,sizeof(unsigned long));
}

void fajl::printinfo(char * psfile)
{
int psifd=open(psfile,O_RDONLY,0600);
	if(psifd<0) throw "open(.psinfo)";

psihdr psih;
int r=read(psifd,&psih,sizeof(psihdr));
	if(r<0) throw "read()";

	if(r!=sizeof(psihdr)){
	Iface->msg(2, "Short read!\n");
	throw 0;
	}

Iface->msg(2, "%s:\n", psfile);
Iface->msg(2, "Size: %lu kb, downloaded in %d parts. Timestamp: %s\n", 
		psih.size/1024, psih.parts,psih.stamp);
fsize_t pmax = (unsigned long)(psih.size/psih.parts) +1;
fsize_t total=0;

	for(int i=0;i<psih.parts;i++){
	fsize_t psize;
		if(read(psifd,&psize,sizeof(psize))!=sizeof(psize)){
		Iface->msg(2, "Short read!\n");
		throw 0;
		}
	total+=psize;
		
		if(i+1 == psih.parts)
		pmax = psih.size - ((psih.parts-1) * pmax);
		
	psize>>=10;
	pmax>>=10;
	Iface->msg(2, "%2d. %8lu kb (%3d%%)\n",i+1,psize,
				(int)((100*psize)/pmax));
	}

total>>=10;
Iface->msg(2, "Tot.%8lu kb (%3d%%)\n",total,(int)((100*total)/(psih.size>>10)));
}

/***/

