#ifndef _THREAD_H
#define _THREAD_H

#include <vector>
#include <unistd.h>
#include <sys/types.h>

using std::vector;

typedef unsigned long fsize_t;

class fajl{
private:
	char * source;		// host
	char * name;		// remote filename
	char * localname;	// local filename

	char psinfo[1024];	// .psinfo filename

	struct psihdr {
		char    stamp[14];
		int     parts;
		unsigned long size;
	};
	
public:
	static vector<char *> sshargs;
	
	// file descriptors
	int fd;
	int psifd;

	char mnglname[1024];	// escaped remote filename
	
	// just like in the header:
	char stamp[14];	
	int parts;
	fsize_t size;

	// size of a typical (not the last) block
	fsize_t block;
	
	fsize_t totalbytes,
		totinitbytes;

	bool force, recover;	// -f & -c
	
	// check remote file size & timestamp
	void get_info();

	// .psinfo file operations
	void writepsinfo();	// write status to file
	void initpsinfo();	// read from file
	static void printinfo(char * psfile);	// print the contents
	void rmpsinfo(){ unlink(psinfo); }

	void setup(char * nm, char * src);

	int spawn_ssh(int *pty,int stout, int sterr, char *cmd, bool compress = false);

/*
   .psinfo file structure:
   [stamp] 14 bytes 			\
   [parts] sizeof(int)=4 bytes 		|----- [psihdr] sizeof(psihdr)=22 bytes
   [size] sizeof(long)=4 bytes 		/
   parts * [bytes] sizeof(long)=4 bytes  (* parts)
 */
};

class thread{
	friend void fajl::initpsinfo();
	friend int main(int, char **);

private:
//		---- static ----
	// select() stuff
	static int maxfd;
	static fd_set fds;
	
public:
	// configuration:
	static bool compress;	// use ssh compression
	static char * proserv;

	// sshs' stderr pipe and password
	static int errpipe[2];
	static char password[100];
	
//		---- non-static ----
	
	int masterpty;	// tty (for password)
	int pipe_fd;	// stdout - reading side
	
//	int transfer;		// 5-s transfer

private:
	
	fsize_t offset,		// position of this block in the file
//		lastbytes,	// FIXME
		ignorebytes;	// amount of data to ignore (for 1kb-boundary)

public:
	fsize_t bytes, 		// amount of data transferred
		initbytes, 	// amount of data we had at the beginning
		max;		// block length
	
	enum { connecting, transferring, finished } status;

	void connect();
//	void recalc-transfer(){transfer=bytes-lastbytes; lastbytes=bytes;
//		fulltransfer+=transfer;};

	static void init();
	void init(int my_num);
	void receive();		// receive pending data

private:
	void remove();		// remove from select() bitfield
				// and close file descriptors
	bool inline is_finished(){return bytes==max;};
};

/***/

extern fajl fle;
extern thread * threads;

#endif
